package com.example

import java.util.Base64
import com.google.auth.oauth2.GoogleCredentials
import com.google.firebase.FirebaseApp
import com.google.firebase.FirebaseOptions
import io.ktor.server.application.*
import io.ktor.server.request.*
import io.ktor.server.response.*
import java.security.MessageDigest
import io.ktor.server.routing.*
import io.ktor.serialization.kotlinx.json.*
import com.google.firebase.messaging.FirebaseMessaging
import com.google.firebase.messaging.Message
import io.ktor.server.engine.*
import kotlinx.serialization.Serializable
import kotlinx.serialization.json.Json
import kotlinx.serialization.json.jsonObject
import kotlinx.serialization.json.jsonPrimitive
import okhttp3.*
import org.rocksdb.Options
import org.rocksdb.RocksDB
import org.rocksdb.RocksDBException
import org.slf4j.LoggerFactory
import java.io.File
import java.io.FileInputStream
import java.io.IOException
import java.security.KeyStore
import kotlinx.coroutines.delay
import kotlinx.coroutines.runBlocking
import org.slf4j.event.Level
import org.slf4j.Logger
import java.nio.file.Files
import java.nio.file.Paths
import io.ktor.server.plugins.defaultheaders.*
import io.ktor.server.plugins.calllogging.*
import io.ktor.server.plugins.contentnegotiation.*
import io.ktor.http.HttpStatusCode
import io.ktor.http.ContentType
import io.ktor.server.engine.embeddedServer
import io.ktor.server.netty.Netty
import io.ktor.server.engine.sslConnector

// For application-level logging
val log: Logger = LoggerFactory.getLogger("Application")

// Represents the request to create an Android Wi-Fi profile
@Serializable
data class AndroidProfileRequest(
    val friendlyName: String,
    val fqdn: String,
    val realm: String,
    val username: String,
    val password: String,       
    val eapType: Int,           // EAP type (e.g., EAP-TTLS)
    val innerMethod: String,    // Inner authentication method (e.g., PAP)     
)

// Represents an authorization request with an authentication code and FCM token
@Serializable
data class AuthRequest(
    val authCode: String,       // OAuth authorization code
    val fcmToken: String,       // Firebase Cloud Messaging token to communicate with mobile app
)

// Represents the response after successful authentication
@Serializable
data class AuthResponse(
    val access_token: String,    // OAuth authorization code
    val refresh_token: String?,  // OAuth refresh token
    val id_token: String,        // ID token
)

// Represents a request with authentication tokens to authenticate user
@Serializable
data class AuthTokens(
    val idToken: String,
    val accessToken: String
)

/*
@Serializable
data class AuthTokensWithRefresh(
    val idToken: String,
    val accessToken: String,
    val refreshToken: String?
)
*/


// To handle interactions with RocksDB
object RocksDBStore {
    private val logger = LoggerFactory.getLogger(RocksDBStore::class.java)
    private var db: RocksDB
    private val options = Options().setCreateIfMissing(true) // Only if it doesn't exist yet
    private const val dbPath = "data/rocksdb"

    init {
        RocksDB.loadLibrary()
        val dbFile = File(dbPath)
        // Create the DB directory if it doesn't exist
        if (!dbFile.exists()) {
            dbFile.mkdirs()
        }
        try {
            db = RocksDB.open(options, dbPath)
            logger.info("RocksDB initialized at $dbPath")
        } catch (e: RocksDBException) {
            logger.error("Error initializing RocksDB: ${e.message}")
            throw IllegalStateException("Failed to initialize RocksDB", e)
        }
    }

    /*
        Stores a key-value pair in the database.

        @param key The key to store.

        @param value The value to associate with the key.
     */
    fun put(key: String, value: String) {
        db.put(key.toByteArray(), value.toByteArray())
    }

    /*
        Retrieves a value by its key.

        @param key The key to look up.
        
        @return The value associated with the key, or null if not found.
     */
    fun get(key: String): String? {
        return db.get(key.toByteArray())?.toString(Charsets.UTF_8)
    }

    /*
        Deletes a key-value pair from the database.
        
        @param key The key to delete.
     */
    fun delete(key: String) {
        db.delete(key.toByteArray())
    }

    /*
        Closes the RocksDB instance and releases resources.
     */
    fun close() {
        options.close()
        db.close()
    }
}

// Application module setup
fun Application.module() {
    install(DefaultHeaders)
    install(CallLogging) {
        level = Level.INFO
    }
    install(ContentNegotiation) {
        json()
    }
    initializeFirebase()

    val log = LoggerFactory.getLogger("Application")

    routing { // HTTPS endpoints definition

        get("/generateAndroidProfile") { // generateAndroidProfile endpoint
        /* 
            A mobile application requested an Hotspot 2.0 profile with its credentials
            This endpoint creates a profile according to the Passpoint (Hotspot 2.0)
            documentation 
            (https://source.android.com/docs/core/connect/wifi-passpoint#passpoint_r2_provisioning)

            Request:
                Content Type: application/x-www-form-urlencoded
                Query Parameters: 
                        friendlyName (String),
                        username (String),
                        password (String)
                    
            Response:
            1   Success (HTTP 200 OK)
                Body: (contains the Hotspot 2.0 profile (Passpoint profile) and the certificate)
                    <Base64-encoded multipart/mixed content>

            2   Error (HTTP 400 Bad Request)
        */
            log.info("Received /generateAndroidProfile request")
            // Extract user specific parameters from the request
            val friendlyName = call.request.queryParameters["friendlyName"]
            val fqdn = "marie.tiedie.io"
            val realm = "test-beid.openroaming.net"
            val username = call.request.queryParameters["username"]
            val to_encode_password = call.request.queryParameters["password"]
            // Encode the password (Documentation)
            val password = Base64.getEncoder().encodeToString(to_encode_password!!.toByteArray(Charsets.UTF_8))
            // EAP configuration
            val eapType = 21 // Corresponds to EAP-TTLS
            val innerMethod = "PAP"
        
            // mandatory parameters
            if (friendlyName == null || username == null || password == null) {
                call.respond(HttpStatusCode.BadRequest, "Missing required parameters")
                return@get
            }
            
            // the Android profile request object
            val profileRequest = AndroidProfileRequest(
                friendlyName = friendlyName,
                fqdn = fqdn,
                realm = realm,
                username = username,
                password = password,
                eapType = eapType,
                innerMethod = innerMethod
            )
        
            // Generate the Passpoint profile and Base64 encode it
            val profileXml = generateAndroidProfileXML(profileRequest)
            val encodedProfile = Base64.getEncoder().encodeToString(profileXml.toByteArray(Charsets.UTF_8))
            
            // Load and encode the CA certificate (for the radius server)
            val caCertPath = "/etc/freeradius/openroaming/auth-server/certs/ca.pem"      
            val encodedCaCert = loadAndEncodeDERCertificate(caCertPath)     

            // Define a boundary (Documentation)
            val boundary = "f6d6201be73d4e46988f789237cffb00"
        
            // Construct the multipart response (Documentation)
            val finalXml = """
                Content-Type: multipart/mixed; boundary=$boundary
                Content-Transfer-Encoding: base64

                --$boundary
                Content-Type: application/x-passpoint-profile
                Content-Transfer-Encoding: base64

                $encodedProfile

                --$boundary
                Content-Type: application/x-x509-ca-cert
                Content-Transfer-Encoding: base64
        
                $encodedCaCert
                --$boundary--
            """.trimIndent()

            // Encode the whole final multipart response in Base64
            val finalBase64Encoded = Base64.getEncoder().encodeToString(finalXml.toByteArray(Charsets.UTF_8))

            // send the response back
            call.respondText(
                finalBase64Encoded,
                ContentType.parse("multipart/mixed; boundary=$boundary"),
                HttpStatusCode.OK
            )
        }
        

        post("/auth") { // auth endpoint
        /* 
            A user successfully authenticated with Google credentials
            via the mobile application. The mobile application sent a /auth 
            request to exchange its Authorization code for access, refresh, id
            tokens.

            Request:
                Content Type: application/json
                Body: {
                        "authCode": "<authorization_code>",
                        "fcmToken": "<fcm_token>"
                    }
            Response:
            1   Success (HTTP 200 OK)
                Body: {
                        "idToken": "<id_token>",
                        "accessToken": "<access_token>",
                        "refreshToken": "<refresh_token>"
                    }

            2   Error (HTTP 400 Bad Request)
            3   Error (HTTP 500 Internal Server Error)
        */
            // Retrieve the content type of the request to check it (Json)
            val requestContentType = call.request.contentType()
            // Retrieve the request body 
            val requestBody = call.receiveText()
            if (requestContentType.match(ContentType.Application.Json)) {
                try {
                    // JSON -> AuthRequest object
                    val authRequest = Json.decodeFromString<AuthRequest>(requestBody)
                    log.info("Received /auth request: $authRequest")
                     // Exchange the authorization code for OAuth tokens
                    val tokens = exchangeAuthCodeForTokens(authRequest.authCode)
                    if (tokens != null) {
                        try {
                            // access token, refresh token, and FCM token must not be null
                            val accessToken = tokens.access_token.takeIf { it.isNotBlank() }
                                ?: throw IllegalArgumentException("Access token is missing or blank")
        
                            val refreshToken = tokens.refresh_token?.takeIf { it.isNotBlank() }
                                ?: throw IllegalArgumentException("Refresh token is missing or blank")

                            val fcmToken = authRequest.fcmToken.takeIf { it.isNotBlank() }
                                ?: throw IllegalArgumentException("FCM token is missing or blank")

                            // accessToken::refreshToken::FMCToken
                            val value = listOfNotNull(
                                accessToken,
                                refreshToken,
                                fcmToken
                            ).joinToString("::")

                            // Extract the "sub" field from the ID token (sub = UID)
                            // sub, The subject: the ID that represents the principal making the request.
                            // From https://cloud.google.com/docs/authentication/token-types
                            val sub = extractSubField(tokens.id_token) ?: throw IllegalArgumentException("Invalid token: 'sub' field is missing")
                            RocksDBStore.put(sub, value)

                            // modifies the response sent back to the client 
                            //(he send id token, we extracted sub from id token)
                            val modifiedTokens = tokens.copy(id_token = sub)
                            call.respond(HttpStatusCode.OK, modifiedTokens)
                            
                        } catch (e: Exception) {
                            log.error("Failed to store tokens: ${e.localizedMessage}")
                            e.printStackTrace()
                            call.respond(HttpStatusCode.InternalServerError, "Failed to store tokens: ${e.localizedMessage}")
                        }
                    } else {
                        call.respond(HttpStatusCode.InternalServerError, "Failed to exchange authorization code for tokens")
                    }
                } catch (e: Exception) {
                    log.error("Failed to process /auth request: ${e.localizedMessage}")
                    e.printStackTrace()
                    call.respond(HttpStatusCode.InternalServerError, "Failed to process /auth request: ${e.localizedMessage}")
                }
            } else {
                call.respond(HttpStatusCode.UnsupportedMediaType, "Content type must be application/json")
            }
        }

        post("/authenticate") { // authenticate endpoint
        /* 
            A user used its Hotspot 2.0 profile to connect to a Wi-Fi network.
            The RADIUS server sent a /authenticate 
            request to ensure that the user is valid and authenticated

            Request:
                Content Type: application/json
                Body: {
                        "idToken": "<id_token>",
                        "accessToken": "<access_token>"
                    }
            Response:
            1   Success (HTTP 200 OK)
                (No body, only a status code indicating successful authentication)

            2   Error (HTTP 401 Unauthorized)
            3   Error (HTTP 400 Bad Request)
            4   Error (HTTP 500 Internal Server Error)
        */
            val requestContentType = call.request.contentType()
            val requestBody = call.receiveText()

            // Ensure the content type is application/json
            if (requestContentType.match(ContentType.Application.Json)) {

                // Deserialize the request body into AuthTokens object
                val authTokens = Json.decodeFromString<AuthTokens>(requestBody)
                log.info("Received /authenticate request: $authTokens")

                // Validate the authentication tokens
                val isAuthenticated = authenticateTokens(authTokens.idToken, authTokens.accessToken)

                if (isAuthenticated) {
                    log.info("Authentication successful for ID Token: ${authTokens.idToken}")
                    call.respond(HttpStatusCode.OK)
                } else {
                    log.info("Authentication failed for ID Token: ${authTokens.idToken}")
                    call.respond(HttpStatusCode.Unauthorized)
                }
            } else {
                call.respond(HttpStatusCode.UnsupportedMediaType, "Content type must be application/json")
            }
        }

        /*post("/addUser") {
            val requestContentType = call.request.contentType()
            val requestBody = call.receiveText()
            if (requestContentType.match(ContentType.Application.Json)) {
                val authTokens = Json.decodeFromString<AuthTokensWithRefresh>(requestBody)
                log.info("Received /addUser request: $authTokens")
                if (RocksDBStore.get(authTokens.idToken) != null) {
                    call.respond(HttpStatusCode.Conflict, "User already exists")
                } else {
                    val value = listOfNotNull(authTokens.accessToken, authTokens.refreshToken).joinToString("::")
                    RocksDBStore.put(authTokens.idToken, value)
                    call.respond(HttpStatusCode.Created, "User added")
                }
            } else {
                call.respond(HttpStatusCode.UnsupportedMediaType, "Content type must be application/json")
            }
        }

        put("/updateUser") {
            val requestContentType = call.request.contentType()
            val requestBody = call.receiveText()
            if (requestContentType.match(ContentType.Application.Json)) {
                val authTokens = Json.decodeFromString<AuthTokensWithRefresh>(requestBody)
                log.info("Received /updateUser request: $authTokens")
                if (RocksDBStore.get(authTokens.idToken) == null) {
                    call.respond(HttpStatusCode.NotFound, "User not found")
                } else {
                    val value = listOfNotNull(authTokens.accessToken, authTokens.refreshToken).joinToString("::")
                    RocksDBStore.put(authTokens.idToken, value)
                    call.respond(HttpStatusCode.OK, "User updated")
                }
            } else {
                call.respond(HttpStatusCode.UnsupportedMediaType, "Content type must be application/json")
            }
        }

        delete("/deleteUser") {
            val requestContentType = call.request.contentType()
            val requestBody = call.receiveText()
            if (requestContentType.match(ContentType.Application.Json)) {
                val authTokens = Json.decodeFromString<AuthTokens>(requestBody)
                log.info("Received /deleteUser request: $authTokens")
                if (RocksDBStore.get(authTokens.idToken) == null) {
                    call.respond(HttpStatusCode.NotFound, "User not found")
                } else {
                    RocksDBStore.delete(authTokens.idToken)
                    call.respond(HttpStatusCode.OK, "User deleted")
                }
            } else {
                call.respond(HttpStatusCode.UnsupportedMediaType, "Content type must be application/json")
            }
        }*/
    }
}

/*
    Initializes the Firebase SDK
    Code taken from the firebase console - service accounts
    https://console.firebase.google.com/project/openrroaming/settings/serviceaccounts/adminsdk
    the serviceAccount contains the service Account Key file in json format
 */
fun initializeFirebase() {
    // Documentation for FirebaseOptions.Builder() : https://firebase.google.com/docs/admin/setup#java_1
    val serviceAccount = FileInputStream("certs/openrroaming-firebase-adminsdk-zxqwv-9005c1ae17.json")
    val options = FirebaseOptions.Builder()
        .setCredentials(GoogleCredentials.fromStream(serviceAccount))
        .setDatabaseUrl("https://openrroaming-default-rtdb.firebaseio.com")
        .build()
    FirebaseApp.initializeApp(options)
}


/*
    Hashes the provided token using SHA-256 and encodes the resulting hash in Base64.

    @param token The string token that needs to be hashed.
    
    @return A Base64-encoded string representing the SHA-256 hash of the token.
 */
fun hashToken(token: String): String {
    return Base64.getEncoder().encodeToString(MessageDigest.getInstance("SHA-256").digest(token.toByteArray()))
}


/*
    Checks if the access token provided is in the database and valid

    @param idToken The id token used (the key for the DB).
    @param accessToken The access token 

    @return true if the provided access token is valid and false otherwise.
 */
fun authenticateTokens(idToken: String, accessToken: String): Boolean {
    // When transmitted, the + become a space but not the case in RocksDB
    val correctedToken = accessToken.replace(" ", "+") 
    // Retrieve the stored value from RocksDB
    val storedValue = RocksDBStore.get(idToken) ?: return false
    val tokens = storedValue.split("::")
    // accessToken::refreshToken::FMCToken so tokens(0)
    val storedAccessToken = tokens.getOrNull(0) ?: return false

    if (storedAccessToken.isEmpty()) {
        log.error("Stored access token is empty or null")
        return false
    }
    // Check if the access token is valid and not expired
    if (!isAccessTokenValid(storedAccessToken)) { 
        log.error("Access token is expired or invalid")
        val refreshToken = tokens.getOrNull(1)
        // Only asks for the new access token if the user provided the right (but expired) access token
        if(refreshToken != null && correctedToken == hashToken(storedAccessToken) && tokens.getOrNull(2) == null){
            val newAccessToken = exchangeRefreshTokenForNewAccessToken(refreshToken)
            val value = listOfNotNull(newAccessToken, refreshToken, tokens.getOrNull(2)).joinToString("::")
            // Update DB
            RocksDBStore.put(idToken, value)

            // Notify device that access token was refreshed using fmcToken
            val deviceToken = tokens.getOrNull(2)
            // Create a message
            val message = Message.builder()
                .setToken(deviceToken)
                .putData("idToken", idToken)
                .putData("accessToken", newAccessToken)
                .build()

            // Send the message
            val response = FirebaseMessaging.getInstance().send(message)
            println("Successfully sent message: $response")
            return false

        }
        return false
    }
    // compare access token provided with the hashed access token from DB
    return correctedToken == hashToken(storedAccessToken)
}



/*
    Checks if the access token provided is valid and not
    expired (via the Google endpoint)

    @param accessToken The access token to validate.
    
    @return true if the token is valid and not expired, false otherwise.
*/
fun isAccessTokenValid(accessToken: String): Boolean {
    val client = OkHttpClient()

    // Build the request
    val request = Request.Builder()
        .url("https://oauth2.googleapis.com/tokeninfo?access_token=$accessToken") //Token info endpoint from Google
        .build()

    return try {
        client.newCall(request).execute().use { response ->
            if (response.isSuccessful) {
                val responseBody = response.body?.string() ?: return false
                val jsonResponse = Json.parseToJsonElement(responseBody).jsonObject

                // Extract the "expires_in" field from the Google response
                val expiresIn = jsonResponse["expires_in"]?.jsonPrimitive?.content?.toIntOrNull()
                // True if it expires in more than 0 sec
                expiresIn != null && expiresIn > 0
            } else {
                log.error("Tokeninfo request failed with code ${response.code}")
                false
            }
        }
    } catch (e: Exception) {
        log.error("Error validating access token: ${e.message}")
        false
    }
}


/*
    Exchange an authorization code provided for id, access
    and refresh tokens via the Google endpoint

    @param authCode The authorization code received from the authorization server
    
    @return An [AuthResponse] containing the tokens if the exchange is successful and null otherwise
*/
fun exchangeAuthCodeForTokens(authCode: String): AuthResponse? {

    val client = OkHttpClient.Builder()
        .build()
    // Load the credentials from the JSON file
    val serviceAccount = File("certs/openrroaming-firebase-adminsdk-zxqwv-9005c1ae17.json").readText()
    val json = Json.parseToJsonElement(serviceAccount).jsonObject

    val clientId = json["client_id"]!!.jsonPrimitive.content
    val clientSecret = json["client_secret"]!!.jsonPrimitive.content
    val redirectUri = "https://marie.tiedie.io/auth" 
    val tokenUrl = json["token_uri"]!!.jsonPrimitive.content

    // as defined in Documentation
    // https://developers.google.com/identity/protocols/oauth2/web-server
    val requestBody = FormBody.Builder()
        .add("code", authCode)
        .add("client_id", clientId)
        .add("client_secret", clientSecret)
        .add("redirect_uri", redirectUri)
        .add("grant_type", "authorization_code")
        .build()

    /*log.info("Request Parameters:")
    log.info("code: $authCode")
    log.info("client_id: $clientId")
    log.info("client_secret: $clientSecret")
    log.info("redirect_uri: $redirectUri")
    log.info("grant_type: authorization_code")
    log.info("Request Body: $requestBody")*/

    val request = Request.Builder()
        .url(tokenUrl)
        .post(requestBody)
        .build()

    log.info("request: $request")

    // Retry mechanism with exponential backoff in case the request didn't work
    val maxRetries = 5
    var attempt = 0
    var backoff = 100L // Initial backoff delay in milliseconds
    while (attempt < maxRetries) {
        try {
            client.newCall(request).execute().use { response ->
                val responseBody = response.body?.string() ?: throw IOException("Empty response body")
                log.info("Google API Response: $responseBody")
                if (response.isSuccessful) {
                    val jsonParser = Json {
                        ignoreUnknownKeys = true // Ignore unknown keys from json
                    }
                    val tokens = jsonParser.decodeFromString<AuthResponse>(responseBody)
                    log.info("Attempt $attempt: Success to exchange auth code for tokens: $responseBody")
                    return AuthResponse(tokens.access_token, tokens.refresh_token, tokens.id_token)
                } else {
                    log.error("Attempt $attempt: Failed to exchange auth code for tokens: $responseBody")
                    throw IOException("Unexpected code $response")
                }
            }
        } catch (e: IOException) {
            log.error("Attempt $attempt: Failed to exchange auth code for tokens: ${e.message}")
            e.printStackTrace() // If issue
            if (attempt < maxRetries - 1) {
                log.info("Retrying in ${backoff / 1000} seconds...")
                runBlocking { delay(backoff) }
                backoff *= 2 // Exponential backoff
            } else {
                return null
            }
        }
        attempt++
    }
    return null
}


/*
    Exchange a refresh token provided for a new, 
    refreshed access token

    @param refreshToken The refresh token provided to obtain a new access token.

    @return The new access token if the exchange is successful and null otherwise
*/

fun exchangeRefreshTokenForNewAccessToken(refreshToken: String): String? {
    val client = OkHttpClient.Builder()
        .build()

    // Load the credentials from the JSON file
    val serviceAccount = File("certs/openrroaming-firebase-adminsdk-zxqwv-9005c1ae17.json").readText()
    val json = Json.parseToJsonElement(serviceAccount).jsonObject

    val clientId = json["client_id"]!!.jsonPrimitive.content
    val clientSecret = json["client_secret"]!!.jsonPrimitive.content
    val tokenUrl = json["token_uri"]!!.jsonPrimitive.content

    //as defined in the documentation
    // https://developers.google.com/identity/protocols/oauth2/web-server#creatingclient
    val requestBody = FormBody.Builder()
        .add("client_id", clientId)
        .add("client_secret", clientSecret)
        .add("grant_type", "refresh_token")
        .add("refresh_token", refreshToken)
        .build()

    val request = Request.Builder()
        .url(tokenUrl)
        .post(requestBody)
        .build()

    //log.info("Request: $request")

    // Retry mechanism with exponential backoff in case the request didn't work
    val maxRetries = 5
    var attempt = 0
    var backoff = 100L // Initial backoff delay in milliseconds
    while (attempt < maxRetries) {
        try {
            client.newCall(request).execute().use { response ->
                val responseBody = response.body?.string() ?: throw IOException("Empty response body")
                if (response.isSuccessful) {
                    val tokens = Json.parseToJsonElement(responseBody).jsonObject
                    val newAccessToken = tokens["accessToken"]?.jsonPrimitive?.content
                    log.info("Attempt $attempt: Successfully refreshed access token: $responseBody")
                    return newAccessToken
                } else {
                    log.error("Attempt $attempt: Failed to refresh access token: $responseBody")
                    throw IOException("Unexpected code $response")
                }
            }
        } catch (e: IOException) {
            log.error("Attempt $attempt: Failed to refresh access token: ${e.message}")
            e.printStackTrace() 
            if (attempt < maxRetries - 1) {
                log.info("Retrying in ${backoff / 1000} seconds...")
                runBlocking { delay(backoff) }
                backoff *= 2 // Exponential backoff
            } else {
                return null
            }
        }
        attempt++
    }
    return null
}


/*
    Creates a Hotspot 2.0 profile by adding some fields needed for the
    user to connect seamlessly to any Wi-Fi network

    @param request An AndroidProfileRequest containing the required fields for the profile.

    @return A string containing the generated XML profile.
    
*/
fun generateAndroidProfileXML(request: AndroidProfileRequest): String {

    val templatePath = Paths.get("src/main/resources/android_profile_template.xml") //the template from documentation
    val template = Files.readString(templatePath)
    return template
        .replace("\${friendlyName}", request.friendlyName)
        .replace("\${fqdn}", request.fqdn)
        .replace("\${realm}", request.realm)
        .replace("\${username}", request.username)
        .replace("\${password}", request.password)
        .replace("\${eapType}", request.eapType.toString())
        .replace("\${innerMethod}", request.innerMethod)
}



/*
    Reads a DER-encoded certificate file and converts its binary content to a Base64-encoded string
    This is needed to include it in the Hotspot 2.0 profile

    @param filePath The path to the DER certificate file.

    @return A Base64-encoded string representation of the certificate content
 */
fun loadAndEncodeDERCertificate(filePath: String): String {
    val certFile = Paths.get(filePath).toFile()
    if (!certFile.exists()) {
        throw IllegalArgumentException("Certificate file not found at $filePath")
    }

    // binary DER 
    val derContent = certFile.readBytes()

    // Encode to Base64
    return Base64.getEncoder().encodeToString(derContent)
}


/*
    Extract the Sub field of a JWT token which 
    corresponds to the UID
    sub, The subject: the ID that represents the principal making the request.
    From https://cloud.google.com/docs/authentication/token-types

    @param jwtToken The JWT from which the "sub" field will be extracted.
    
    @return The value of the "sub" field if it exists, or null otherwise

*/
fun extractSubField(jwtToken: String): String? {
    // Split the JWT : header.payload.signature
    val parts = jwtToken.split(".") 
    if (parts.size < 2) return null // Invalid JWT format

    // Decode the payload (second part of the JWT)
    val payload = String(Base64.getUrlDecoder().decode(parts[1]))

    // Find and return the "sub" field (= UID of user)
    val json = Json.parseToJsonElement(payload).jsonObject
    val sub = json["sub"]?.jsonPrimitive?.takeIf { it.isString }?.content
    return sub
    
} 

fun main() {
    embeddedServer(Netty, applicationEnvironment { log = LoggerFactory.getLogger("ktor.application") }, {
        envConfig()
    }, module = Application::module).start(wait = true)
}

private fun ApplicationEngine.Configuration.envConfig() {

    // PKCS#12 format -> keystore
    val keyStoreFile = File("certs/server.p12")
    val keyStorePassword = "".toCharArray() 
    val keyStore = KeyStore.getInstance("PKCS12").apply {
        keyStoreFile.inputStream().use { load(it, keyStorePassword) }
    }

    sslConnector(
        keyStore = keyStore,
        keyAlias = "ktorServer",
        keyStorePassword = { keyStorePassword },
        privateKeyPassword = { keyStorePassword }) {
        port = 443 //HTTPS
        keyStorePath = keyStoreFile
    }
}