package com.example.firebaseauth.ui.theme

import android.content.Context
import android.net.ConnectivityManager
import android.net.NetworkCapabilities
import android.app.Activity
import android.os.Build
import android.util.Log
import android.widget.Toast
import androidx.activity.compose.rememberLauncherForActivityResult
import androidx.activity.result.contract.ActivityResultContracts
import androidx.annotation.RequiresApi
import androidx.compose.foundation.layout.*
import androidx.compose.material3.Button
import androidx.compose.material3.Text
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.platform.LocalContext
import androidx.compose.ui.unit.dp
import com.example.firebaseauth.R
import com.example.firebaseauth.auth.AuthState
import com.example.firebaseauth.auth.AuthViewModel
import com.google.android.gms.auth.api.signin.GoogleSignIn
import com.google.android.gms.auth.api.signin.GoogleSignInOptions
import com.google.android.gms.common.api.Scope
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.launch
import androidx.compose.foundation.Image
import androidx.compose.foundation.shape.RoundedCornerShape
import androidx.compose.runtime.Composable
import androidx.compose.runtime.collectAsState
import androidx.compose.runtime.getValue
import androidx.compose.ui.graphics.Color
import androidx.compose.ui.res.painterResource
import androidx.compose.ui.text.style.TextAlign
import androidx.compose.ui.unit.sp
import com.example.firebaseauth.utils.AndroidProfileRequest
import com.example.firebaseauth.utils.sendAuthCodeToServer
import com.example.firebaseauth.utils.sendGenerateAndroidProfileRequest
import java.io.IOException
import java.security.MessageDigest
import java.util.Base64


fun isNetworkAvailable(context: Context): Boolean {
    // Function to check network availability (sign in won't work without connectivity)
    val connectivityManager = context.getSystemService(Context.CONNECTIVITY_SERVICE) as ConnectivityManager
    val activeNetwork = connectivityManager.activeNetwork
    val networkCapabilities = connectivityManager.getNetworkCapabilities(activeNetwork)
    return networkCapabilities != null &&
            networkCapabilities.hasCapability(NetworkCapabilities.NET_CAPABILITY_INTERNET)
}

@RequiresApi(Build.VERSION_CODES.O)
@Composable
fun AuthScreen(viewModel: AuthViewModel) {
    val authState by viewModel.authState.collectAsState() // user authenticated or not
    val activity = LocalContext.current as Activity
    val context = LocalContext.current


    // Launcher for Google Sign-In activity
    val launcher = rememberLauncherForActivityResult(ActivityResultContracts.StartActivityForResult()) { result ->
        val task = GoogleSignIn.getSignedInAccountFromIntent(result.data)
        val account = task.result
        account?.idToken?.let { idToken ->
            viewModel.signInWithGoogle(idToken) // Update ViewModel state with Google ID token
            // Retrieve Authorization Code
            val authCode = account.serverAuthCode
            Log.d("AuthScreen", "Authorization Code received")

            // Send the Authorization Code to the backend server
            authCode?.let { _ ->
                CoroutineScope(Dispatchers.IO).launch {
                    try {
                        // Set up request with authorization code for auth server to receive tokens
                        val tokens = sendAuthCodeToServer(context, authCode)
                        val subidToken = tokens?.first // id token (sub field = UID)
                        val accessToken = tokens?.second // access token
                        Log.d("AuthScreen", "subidToken received: $subidToken")

                        if (accessToken != null) {
                            // After successful authentication, create Android profile request
                            // access token needs to be hashed otherwise too long for the profile
                            val hashedAccessToken = Base64.getEncoder().encodeToString(MessageDigest.getInstance("SHA-256").digest(accessToken.toByteArray()))
                            val profileRequest = subidToken?.let {
                                    AndroidProfileRequest(
                                        friendlyName = "OpenRoaming network (e-ID)", // should be google but the goal of the project was eid
                                        username = subidToken,
                                        password = hashedAccessToken)
                            }
                            // Send the profile request to the server
                            if (profileRequest != null) {
                                sendGenerateAndroidProfileRequest(context, profileRequest)
                            }
                        } else {
                            throw IOException("Failed to retrieve tokens")
                        }
                    } catch (e: Exception) {
                        Log.e("AuthScreen", "Failed to retrieve profile", e)
                    }
                }
            }
        }
    }

    Box(
        modifier = Modifier.fillMaxSize(),
        contentAlignment = Alignment.Center
    ) {
        when (authState) { //UI according to the user state
            is AuthState.Loading -> {
                Text("Loading...")
            }
            is AuthState.Authenticated -> {
            }
            is AuthState.Unauthenticated -> { // sign in page
                Column(
                    modifier = Modifier
                        .fillMaxWidth()
                        .padding(16.dp),
                    horizontalAlignment = Alignment.CenterHorizontally,
                    verticalArrangement = Arrangement.SpaceBetween
                ) {
                    // Logo
                    Image(
                        painter = painterResource(id = R.drawable.openroaming),
                        contentDescription = "OpenRoaming Logo",
                        modifier = Modifier.size(300.dp)
                    )

                    Spacer(modifier = Modifier.height(16.dp))

                    // Description
                    Text(
                        text = "As a Belgian citizen, you can use this application to get access to seamless and secure Wifi. Please choose an account" +
                                "\n" +
                                "\n" +
                                "\n" +
                                "\n" +
                                "\n" +
                                "\n" ,
                        fontSize = 18.sp,
                        color = Color.Gray,
                        textAlign = TextAlign.Center
                    )

                    Spacer(modifier = Modifier.height(32.dp))

                    // Sign In Buttons
                    Button(
                        onClick = {
                            if (!isNetworkAvailable(context)) {
                                // Toast to avoid user trying to sign in without connectivity
                                val toast = Toast.makeText(context, "No internet connection. Please connect to a network and try again.", Toast.LENGTH_SHORT)
                                toast.show()
                                val handler = android.os.Handler(android.os.Looper.getMainLooper())
                                handler.postDelayed({
                                    toast.cancel()
                                }, 1500) // show for 1.5 sec

                                return@Button
                            }
                            val gso = GoogleSignInOptions.Builder(GoogleSignInOptions.DEFAULT_SIGN_IN)
                                .requestEmail()
                                .requestProfile()
                                .requestIdToken(context.getString(R.string.web_client_id))
                                .requestServerAuthCode(context.getString(R.string.web_client_id), true)
                                .requestScopes(
                                    Scope("https://www.googleapis.com/auth/userinfo.email"),
                                    Scope("openid"), //manfatory for OIDC
                                    Scope("email")  // email for legal reasons (Cisco)
                                )
                                .build()

                            val googleSignInClient = GoogleSignIn.getClient(activity, gso)
                            launcher.launch(googleSignInClient.signInIntent)
                        },
                        modifier = Modifier.fillMaxWidth(0.8f),
                        shape = RoundedCornerShape(8.dp)
                    ) {
                        Image(
                            painter = painterResource(id = R.drawable.google),
                            contentDescription = "Google Logo",
                            modifier = Modifier.size(30.dp)
                        )
                        Spacer(modifier = Modifier.width(8.dp))
                        Text("Sign In with Google")
                    }

                    Spacer(modifier = Modifier.height(16.dp))

                    Button(
                        onClick = { /* eID login not set up yet! This is a placeholder! */ },
                        modifier = Modifier.fillMaxWidth(0.8f),
                        shape = RoundedCornerShape(8.dp)
                    ) {
                        Image(
                            painter = painterResource(id = R.drawable.eid),
                            contentDescription = "eID Logo",
                            modifier = Modifier.size(30.dp)
                        )
                        Spacer(modifier = Modifier.width(8.dp))
                        Text("Sign In with eID")
                    }

                    Spacer(modifier = Modifier.weight(1f))

                    // Cisco Logo
                    Image(
                        painter = painterResource(id = R.drawable.cisco),
                        contentDescription = "Cisco Logo",
                        modifier = Modifier.size(80.dp)
                    )
                }
            }
        }
    }
}
