package com.example.firebaseauth.ui.theme

import android.app.Activity
import android.util.Log
import androidx.activity.compose.rememberLauncherForActivityResult
import androidx.activity.result.contract.ActivityResultContracts
import androidx.compose.foundation.layout.*
import androidx.compose.material3.Button
import androidx.compose.material3.Scaffold
import androidx.compose.material3.Text
import androidx.compose.runtime.Composable
import androidx.compose.runtime.collectAsState
import androidx.compose.runtime.getValue
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.platform.LocalContext
import androidx.compose.ui.tooling.preview.Preview
import androidx.compose.ui.unit.dp
import androidx.navigation.NavHostController
import androidx.navigation.compose.rememberNavController
import com.example.firebaseauth.auth.AuthRepository
import com.example.firebaseauth.auth.AuthState
import com.example.firebaseauth.auth.AuthViewModel
import com.example.firebaseauth.utils.sendAuthCodeToServer
import com.google.android.gms.auth.api.signin.GoogleSignIn
import com.google.android.gms.auth.api.signin.GoogleSignInOptions
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.launch


@Composable
fun MainPage(navController: NavHostController, authViewModel: AuthViewModel) {
    val authState by authViewModel.authState.collectAsState() // user authenticated or not
    val activity = LocalContext.current as Activity
    val context = LocalContext.current

    val launcher = rememberLauncherForActivityResult(ActivityResultContracts.StartActivityForResult()) { result ->
        // result of the Google sign-in activity
        val task = GoogleSignIn.getSignedInAccountFromIntent(result.data)
        val account = task.result
        account?.idToken?.let { idToken ->
            authViewModel.signInWithGoogle(idToken)

            // Retrieve Authorization Code
            val authCode = account.serverAuthCode
            Log.d("MainPage", "Authorization Code: $authCode")

            // Send the Authorization Code to your backend server
            authCode?.let { code ->
                CoroutineScope(Dispatchers.IO).launch {
                    try {
                        sendAuthCodeToServer(context, code)
                    } catch (e: Exception) {
                        Log.e("MainPage", "Failed to send auth code to server", e)
                    }
                }
            }
        }
    }

    Scaffold(modifier = Modifier.fillMaxSize()) { innerPadding ->
        Box(
            modifier = Modifier
                .fillMaxSize()
                .padding(innerPadding),
            contentAlignment = Alignment.Center
        ) {
            when (authState) { // UI based on auth state
                is AuthState.Authenticated -> {
                    Column(horizontalAlignment = Alignment.CenterHorizontally) {

                        // Display a welcome message with the user's name retrieved from authentication
                        Text("Welcome, to the OpenRoaming app, ${(authState as AuthState.Authenticated).user.displayName} !")
                        Spacer(modifier = Modifier.height(16.dp))
                        Button(onClick = {
                            // Sign out
                            authViewModel.signOut()

                            // Sign out from Google
                            val gso = GoogleSignInOptions.Builder(GoogleSignInOptions.DEFAULT_SIGN_IN).build()
                            val googleSignInClient = GoogleSignIn.getClient(activity, gso)
                            googleSignInClient.signOut().addOnCompleteListener {
                                // Navigate back to the login page
                                navController.navigate("auth") {
                                    popUpTo(navController.graph.startDestinationId) {
                                        inclusive = true
                                    }
                                }
                            }
                        }) {
                            Text("Sign Out")
                        }
                    }
                }
                is AuthState.Unauthenticated -> {
                    Column(horizontalAlignment = Alignment.CenterHorizontally) {
                        Text("Please sign in to continue.")
                    }
                }
                is AuthState.Loading -> {
                    Text("Checking authentication status...")
                }
            }
        }
    }
}

@Preview(showBackground = true)
@Composable
fun MainPagePreview() {
    val navController = rememberNavController()
    val authViewModel = AuthViewModel(AuthRepository())
    FireBaseAuthTheme {
        MainPage(navController, authViewModel)
    }
}
