package com.example.firebaseauth.utils

import android.annotation.SuppressLint
import android.content.Context
import android.util.Log
import okhttp3.MediaType.Companion.toMediaType
import okhttp3.RequestBody.Companion.toRequestBody
import org.json.JSONObject
import java.io.IOException
import android.net.wifi.WifiManager
import android.net.wifi.WifiNetworkSuggestion
import android.net.wifi.hotspot2.ConfigParser
import com.example.firebaseauth.R
import com.google.firebase.messaging.FirebaseMessaging
import okhttp3.OkHttpClient
import okhttp3.Request
import java.security.KeyStore
import java.security.SecureRandom
import java.security.cert.CertificateFactory
import java.security.cert.X509Certificate
import javax.net.ssl.SSLContext
import javax.net.ssl.TrustManagerFactory
import javax.net.ssl.X509TrustManager

// Data class representing a request to generate an Android Wi-Fi profile
data class AndroidProfileRequest(
    val friendlyName: String,
    val username: String,
    val password: String
)
@SuppressLint("NewApi")
// Function to create a OkHttpClient with custom CA certificate (auth server ca)
fun getOkHttpClient(context: Context): OkHttpClient {
    val certificateFactory = CertificateFactory.getInstance("X.509")
    val inputStream = context.resources.openRawResource(R.raw.ca) // Reference to ca.pem
    val certificate = inputStream.use {
        certificateFactory.generateCertificate(it) as X509Certificate
    }
    // keystore with the auth server ca
    val keyStore = KeyStore.getInstance(KeyStore.getDefaultType()).apply {
        load(null, null)
        setCertificateEntry("ca", certificate)
    }

    val trustManagerFactory = TrustManagerFactory.getInstance(TrustManagerFactory.getDefaultAlgorithm()).apply {
        init(keyStore)
    }

    val trustManagers = trustManagerFactory.trustManagers
    val sslContext = SSLContext.getInstance("TLS").apply {
        init(null, trustManagers, SecureRandom())
    }

    return OkHttpClient.Builder()
        .sslSocketFactory(sslContext.socketFactory, trustManagers[0] as X509TrustManager)
        .hostnameVerifier { _, _ -> true } //no hostname verification (can be added for security)
        .build()
}

// Function to send an authentication code and FCM token to the backend server
// in order to retrieve a id, access, refresh tokens
fun sendAuthCodeToServer(context: Context, authCode: String): Pair<String, String>? {
    val backendUrl = "https://marie.tiedie.io/auth" // Use HTTPS

    var fcmToken: String? = null //For firebase cloud messaging to know when access token is refreshed
    getFCMToken { token ->
        fcmToken = token
    }
    val client = getOkHttpClient(context)

    // Construct the JSON body to send both authCode and fcmToken
    val json = """{
        "authCode": "$authCode",
        "fcmToken": "$fcmToken"
    }"""
    val mediaType = "application/json".toMediaType()
    val requestBody = json.toRequestBody(mediaType)
    val request = Request.Builder()
        .url(backendUrl)
        .post(requestBody)
        .addHeader("Content-Type", "application/json")
        .addHeader("Accept", "application/json")
        .build()

    try {
        client.newCall(request).execute().use { response ->
            if (!response.isSuccessful) {
                throw IOException("Unexpected response code: ${response.code}")
            }

            val responseBody = response.body?.string()
            if (responseBody.isNullOrEmpty()) {
                throw IOException("Empty response body")
            }

            try {
                val jsonObject = JSONObject(responseBody)
                val accessToken = jsonObject.getString("access_token")
                val idToken = jsonObject.getString("id_token")

                // Return both idToken and accessToken as a pair
                return Pair(idToken, accessToken) // needed for the profile
            } catch (e: Exception) {
                throw Exception("Failed to parse response JSON: $responseBody", e)
            }
        }
    } catch (e: IOException) {
        Log.e("AuthUtils", "Network error: ${e.message}", e)
    } catch (e: Exception) {
        Log.e("AuthUtils", "Unexpected error: ${e.message}", e)
    }

    return null
}



@SuppressLint("NewApi", "ServiceCast")
fun sendGenerateAndroidProfileRequest(
    context: Context,
    profileRequest: AndroidProfileRequest
) {
    val backendUrl = "https://marie.tiedie.io/generateAndroidProfile" // Use HTTPS (auth server endpoint)

    val client = getOkHttpClient(context)

    // Construct the query parameters for the GET request
    val url = "$backendUrl?" +
            "friendlyName=${profileRequest.friendlyName}&" +
            "username=${profileRequest.username}&" +
            "password=${profileRequest.password}&"

    val request = Request.Builder()
        .url(url)  // Use GET request with the parameters in the URL
        .get()  // Explicitly use GET
        .addHeader("Accept", "application/x-wifi-config")  // Expected response type
        .build()

    try {
        client.newCall(request).execute().use { response ->
            if (!response.isSuccessful) {
                throw IOException("Unexpected code $response")
            }

            // Get the response as a byte array
            val respByte = response.body?.bytes()
            if (respByte != null) {
                try {
                    // Now parse the Passpoint configuration
                    val actualConfig = ConfigParser.parsePasspointConfig(
                        "application/x-wifi-config",
                        respByte
                    )

                    if (actualConfig != null) {
                        // Provisioning a PasspointConfiguration is supported by the suggestion API:
                        // https://developer.android.com/develop/connectivity/wifi/wifi-suggest
                        val wifiManager =
                            context.getSystemService(Context.WIFI_SERVICE) as WifiManager
                        val suggestion3 = WifiNetworkSuggestion.Builder()
                            .setPasspointConfig(actualConfig)
                            .setIsAppInteractionRequired(true) // Optional (Needs location permission)
                            .build()

                        val suggestionsList = listOf(suggestion3)

                        val status = wifiManager.addNetworkSuggestions(suggestionsList)
                        if (status != WifiManager.STATUS_NETWORK_SUGGESTIONS_SUCCESS) {
                            Log.d("AuthUtils", "Network suggestion not added")
                        }

                        Log.d("AuthUtils", "Passpoint profile installed successfully")
                    } else {
                        Log.e("AuthUtils", "Failed to parse Passpoint configuration")
                    }
                } catch (e: Exception) {
                    Log.e("AuthUtils", "Error processing Passpoint profile", e)
                }
            } else {
                Log.e("AuthUtils", "Response body is null")
            }
        }
    } catch (e: IOException) {
        Log.e("GenerateProfile", "Network call failed", e)
    }
}

// Function to retrieve the Firebase Cloud Messaging (FCM) token
// from https://firebase.google.com/docs/cloud-messaging/android/client
fun getFCMToken(onTokenReceived: (String?) -> Unit) {
    FirebaseMessaging.getInstance().token.addOnCompleteListener { task ->
        if (task.isSuccessful) {
            // Pass the token to the callback
            onTokenReceived(task.result)
        } else {
            Log.d("FCM", "Fetching FCM registration token failed", task.exception)
            onTokenReceived(null)
        }
    }
}
